#region License

/* Copyright (c) 2007 Leslie Sanford
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to 
 * deal in the Software without restriction, including without limitation the 
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
 * sell copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software. 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
 * THE SOFTWARE.
 */

#endregion

#region Contact

/*
 * Leslie Sanford
 * Email: jabberdabber@hotmail.com
 */

#endregion

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using System.Timers;
using Sanford.Multimedia;
using Sanford.Multimedia.Timers;
using Sanford.Threading;

namespace Sanford.Multimedia.Wave
{
	/// <summary>
	/// Summary description for OutputDevice.
	/// </summary>
	public partial class OutputDevice : WaveDevice
	{
        #region Win32 Wave Constants
        
        private const int WOM_DONE = 0x3BD;

        private const int WAVECAPS_PITCH        = 0x0001;       /* supports pitch control */
        private const int WAVECAPS_PLAYBACKRATE = 0x0002;       /* supports playback rate control */
        private const int WAVECAPS_VOLUME       = 0x0004;       /* supports volume control */
        private const int WAVECAPS_LRVOLUME     = 0x0008;       /* separate left-right volume control */

        private const int WHDR_BEGINLOOP        = 0x00000004;   /* loop start block */
        private const int WHDR_ENDLOOP          = 0x00000008;   /* loop end block */

        private const int CALLBACK_NULL = 0x00000000;

        #endregion
      
        private IntPtr handle = IntPtr.Zero;

        private WaveOutCaps caps = new WaveOutCaps();

        private Queue<OutputAsyncResult> asyncResults = new Queue<OutputAsyncResult>();

        private System.Timers.Timer timer = new System.Timers.Timer();

        private readonly object lockObject = new object();

        public OutputDevice(int deviceID, short channelCount, int samplesPerSecond, short bitsPerSecond) 
            : base(deviceID)
		{
            NativeMethods.waveOutGetDevCaps(deviceID, ref caps, Marshal.SizeOf(typeof(WaveOutCaps)));

            WaveFormat format = CreateWaveFormat(channelCount, samplesPerSecond, bitsPerSecond);

            int result = NativeMethods.waveOutOpen(out handle, deviceID, ref format, IntPtr.Zero, IntPtr.Zero, CALLBACK_NULL);

            if(result != DeviceException.MMSYSERR_NOERROR)
            {
                throw new OutputDeviceException(result);
            }

            timer.Elapsed += ElapsedHandler;

            timer.Start();
        }

        private void ElapsedHandler(object sender, ElapsedEventArgs e)
        {
            lock(lockObject)
            {
                if(asyncResults.Count > 0)
                {
                    OutputAsyncResult asyncResult = asyncResults.Peek();
                    WaveformBuffer buffer = asyncResult.Buffer;

                    if(buffer.IsDone)
                    {
                        int result = NativeMethods.waveOutUnprepareHeader(handle, buffer.HeaderPointer, SizeOfWaveHeader);

                        if(result != OutputDeviceException.MMSYSERR_NOERROR)
                        {
                            asyncResult.Error = new OutputDeviceException(result);
                        }

                        asyncResult.Signal();
                        asyncResults.Dequeue();

                        Monitor.Pulse(lockObject);
                    }
                }
            }
        }

        ~OutputDevice()
        {
            Dispose(false);
        }

        protected override void Dispose(bool disposing)
        {
            if(disposing)
            {
                Reset();

                timer.Dispose();

                int result = NativeMethods.waveOutClose(handle);

                if(result != DeviceException.MMSYSERR_NOERROR)
                {
                    throw new OutputDeviceException(result);
                }
            }
            else
            {
                NativeMethods.waveOutReset(handle);
                NativeMethods.waveOutClose(handle);
            }

            base.Dispose(disposing);
        }

        public override void Close()
        {
            #region Guard

            if(IsDisposed)
            {
                return;
            }

            #endregion            

            Dispose();
        }

        public override void Reset()
        {
            #region Require

            if(IsDisposed)
            {
                throw new ObjectDisposedException("OutputDevice");
            }

            #endregion

            lock(lockObject)
            {
                int result = NativeMethods.waveOutReset(handle);

                if(result == DeviceException.MMSYSERR_NOERROR)
                {
                    while(asyncResults.Count > 0)
                    {
                        Monitor.Wait(lockObject);
                    }
                }
                else
                {
                    throw new OutputDeviceException(result);
                }
            }            
        }

        public void Pause()
        {
            #region Require

            if(IsDisposed)
            {
                throw new ObjectDisposedException("OutputDevice");
            }

            #endregion

            lock(lockObject)
            {
                int result = NativeMethods.waveOutPause(handle);

                if(result != DeviceException.MMSYSERR_NOERROR)
                {
                    throw new OutputDeviceException(result);
                }
            }
        }

        public void Restart()
        {
            #region Require

            if(IsDisposed)
            {
                throw new ObjectDisposedException("OutputDevice");
            }

            #endregion

            lock(lockObject)
            {
                int result = NativeMethods.waveOutRestart(handle);

                if(result != DeviceException.MMSYSERR_NOERROR)
                {
                    throw new OutputDeviceException(result);
                }
            }
        }

        public IAsyncResult BeginSend(WaveformBuffer buffer, AsyncCallback callback, object state)
        {
            #region Require

            if(IsDisposed)
            {
                throw new ObjectDisposedException("OutputDevice");
            }
            else if(buffer == null)
            {
                throw new ArgumentNullException("wave");
            }

            #endregion

            OutputAsyncResult asyncResult = new OutputAsyncResult(buffer, this, callback, state);            

            lock(lockObject)
            {
                buffer.CopyToMemory();

                int result = NativeMethods.waveOutPrepareHeader(handle, buffer.HeaderPointer, SizeOfWaveHeader);

                if(result == DeviceException.MMSYSERR_NOERROR)
                {
                    result = NativeMethods.waveOutWrite(handle, buffer.HeaderPointer, SizeOfWaveHeader);

                    if(result != DeviceException.MMSYSERR_NOERROR)
                    {
                        NativeMethods.waveOutUnprepareHeader(handle, buffer.HeaderPointer, SizeOfWaveHeader);

                        throw new OutputDeviceException(result);
                    }
                }
                else
                {
                    throw new OutputDeviceException(result);
                }

                asyncResults.Enqueue(asyncResult);
            }

            return asyncResult;                                 
        }

        public void EndSend(IAsyncResult result)
        {
            OutputAsyncResult r = result as OutputAsyncResult;

            if(r == null)
            {
                throw new ArgumentException("Wrong type of IAsyncResult object.");
            }
            else if(r.Owner != this)
            {
                throw new ArgumentException("This IAsyncResult object did not originate on this instance.");
            }

            r.AsyncWaitHandle.WaitOne();

            if(r.Error != null)
            {
                throw r.Error;
            }
        }

        public void Send(WaveformBuffer buffer)
        {
            EndSend(BeginSend(buffer, null, null));
        }

        public Time GetPosition(TimeType type)
        {
            #region Require

            if(IsDisposed)
            {
                throw new ObjectDisposedException("OutputDevice");
            }

            #endregion

            Time t = new Time();

            t.type = (int)type;

            int result = NativeMethods.waveOutGetPosition(handle, ref t, Marshal.SizeOf(typeof(Time)));

            if(result != DeviceException.MMSYSERR_NOERROR)
            {
                throw new OutputDeviceException(result);
            }

            return t;
        }

        public static WaveOutCaps GetDeviceCapabilities(int deviceID)
        {
            WaveOutCaps caps = new WaveOutCaps();
            int result = NativeMethods.waveOutGetDevCaps(deviceID, ref caps, Marshal.SizeOf(typeof(WaveOutCaps)));

            if(result != DeviceException.MMSYSERR_NOERROR)
            {
                throw new OutputDeviceException(result);
            }

            return caps;
        }

        private static WaveFormat CreateWaveFormat(short channelCount, int samplesPerSecond, short bitsPerSecond)
        {
            #region Require

            if(channelCount < 1)
            {
                throw new ArgumentOutOfRangeException("channelCount");
            }
            else if(samplesPerSecond < 1)
            {
                throw new ArgumentOutOfRangeException("samplesPerSecond");
            }
            else if(bitsPerSecond < 1)
            {
                throw new ArgumentOutOfRangeException("bitsPerSecond");
            }

            #endregion

            WaveFormat format = new WaveFormat();

            format.FormatTag = 1;   // PCM/Uncompressed.
            format.SizeOfExtraData = 0;
            format.Channels = channelCount;
            format.BitsPerSample = bitsPerSecond;
            format.SamplesPerSecond = samplesPerSecond;

            if(bitsPerSecond % 8 == 0)
            {
                format.BlockAlign = (short)(bitsPerSecond / 8 * channelCount);
            }
            else
            {
                format.BlockAlign = (short)(bitsPerSecond / 8 * channelCount + 1);
            }

            format.AverageBytesPerSeconds = format.BlockAlign * samplesPerSecond;

            return format;
        }

        public bool CanSetPitch
        {
            get
            {
                #region Require

                if(IsDisposed)
                {
                    throw new ObjectDisposedException("OutputDevice");
                }

                #endregion

                return (caps.support & WAVECAPS_PITCH) == WAVECAPS_PITCH;
            }
        }

        public bool CanSetPlaybackRate
        {
            get
            {
                #region Require

                if(IsDisposed)
                {
                    throw new ObjectDisposedException("OutputDevice");
                }

                #endregion

                return (caps.support & WAVECAPS_PLAYBACKRATE) == WAVECAPS_PLAYBACKRATE;
            }
        }

        public bool CanSetVolume
        {
            get
            {
                #region Require

                if(IsDisposed)
                {
                    throw new ObjectDisposedException("OutputDevice");
                }

                #endregion

                return (caps.support & WAVECAPS_VOLUME) == WAVECAPS_VOLUME;
            }
        }

        public bool CanSetStereoVolume
        {
            get
            {
                #region Require

                if(IsDisposed)
                {
                    throw new ObjectDisposedException("OutputDevice");
                }

                #endregion

                return (caps.support & WAVECAPS_LRVOLUME) == WAVECAPS_LRVOLUME;
            }
        }

        public int Pitch
        {
            get
            {
                #region Require

                if(IsDisposed)
                {
                    throw new ObjectDisposedException("OutputDevice");
                }

                #endregion

                int pitch = 0;

                int result = NativeMethods.waveOutGetPitch(handle, ref pitch);

                if(result != DeviceException.MMSYSERR_NOERROR)
                {
                    throw new OutputDeviceException(result);
                }

                return pitch;
            }
            set
            {
                #region Require

                if(value < 0)
                {
                    throw new ArgumentOutOfRangeException("PitchMultiplier", value,
                        "Pitch multiplier out of range.");
                }

                #endregion

                int result = NativeMethods.waveOutSetPitch(handle, value);

                if(result != DeviceException.MMSYSERR_NOERROR)
                {
                    throw new OutputDeviceException(result);
                }
            }
        }

        public int PlaybackRate
        {
            get
            {
                #region Require

                if(IsDisposed)
                {
                    throw new ObjectDisposedException("OutputDevice");
                }

                #endregion

                int rate = 0;

                int result = NativeMethods.waveOutGetPlaybackRate(handle, ref rate);

                if(result != DeviceException.MMSYSERR_NOERROR)
                {
                    throw new OutputDeviceException(result);
                }

                return rate;
            }
            set
            {
                #region Require

                if(IsDisposed)
                {
                    throw new ObjectDisposedException("OutputDevice");
                } 
                else if(value < 0)
                {
                    throw new ArgumentOutOfRangeException("PlaybackRateMultiplier", value,
                        "Playback rate multiplier out of range.");
                }

                #endregion

                int result = NativeMethods.waveOutSetPlaybackRate(handle, value);

                if(result != DeviceException.MMSYSERR_NOERROR)
                {
                    throw new OutputDeviceException(result);
                }
            }
        }

        public int Volume
        {
            get
            {
                #region Require

                if(IsDisposed)
                {
                    throw new ObjectDisposedException("OutputDevice");
                }

                #endregion

                int volume = 0;

                int result = NativeMethods.waveOutGetVolume(handle, ref volume);

                if(result != DeviceException.MMSYSERR_NOERROR)
                {
                    throw new OutputDeviceException(result);
                }

                return volume;
            }
            set
            {
                #region Require

                if(IsDisposed)
                {
                    throw new ObjectDisposedException("OutputDevice");
                }

                #endregion

                int result = NativeMethods.waveOutSetVolume(handle, value);

                if(result != DeviceException.MMSYSERR_NOERROR)
                {
                    throw new OutputDeviceException(result);
                }
            }
        }

        public double PollingInterval
        {
            get
            {
                return timer.Interval;
            }
            set
            {
                timer.Interval = value;
            }
        }

        public static int DeviceCount
        {
            get
            {
                return NativeMethods.waveOutGetNumDevs();
            }
        }

        #region IDisposable Members

        public override void Dispose()
        {
            #region Guard

            if(IsDisposed)
            {
                return;
            }

            #endregion

            Dispose(true);

            GC.SuppressFinalize(this);
        }

        #endregion

        #region Device Members

        public override IntPtr Handle
        {
            get 
            {
                return handle;
            }
        }

        #endregion        
    }

    public class OutputDeviceException : WaveDeviceException
    {
        #region OutputDeviceException Members

        #region Fields

        // The error message.
        private StringBuilder message = new StringBuilder(128);        

        #endregion

        #region Construction

        /// <summary>
        /// Initializes a new instance of the OutputDeviceException class with
        /// the specified error code.
        /// </summary>
        /// <param name="errCode">
        /// The error code.
        /// </param>
        public OutputDeviceException(int errCode) : base(errCode)
        {
            // Get error message.
            NativeMethods.waveOutGetErrorText(errCode, message, message.Capacity);
        }

        #endregion

        #region Properties

        /// <summary>
        /// Gets a message that describes the current exception.
        /// </summary>
        public override string Message
        {
            get
            {
                return message.ToString();
            }
        }        

        #endregion

        #endregion
    }
}
