#region License

/* Copyright (c) 2007 Leslie Sanford
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to 
 * deal in the Software without restriction, including without limitation the 
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
 * sell copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software. 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
 * THE SOFTWARE.
 */

#endregion

#region Contact

/*
 * Leslie Sanford
 * Email: jabberdabber@hotmail.com
 */

#endregion

using System;
using System.Diagnostics;
using System.IO;

namespace Sanford.Multimedia.Wave
{
    /// <summary>
    /// 
    /// </summary>
    public sealed class WaveReader : IDisposable
    {
        #region WaveReader Members

        #region Fields

        #region Constants

        private const int ChunkIDLength = 4;

        #endregion

        private Stream input;

        private short channelCount;

        private int samplesPerSecond;

        private short bitsPerSecond;

        private long waveChunkPosition;

        private long bytesLeft;

        private long waveDataLength;

        private bool disposed = false;

        #endregion

        #region Construction

        public WaveReader(Stream input)
        {
            #region Require

            if(input == null)
            {
                throw new ArgumentNullException("input");
            }
            else if(!input.CanRead)
            {
                throw new ArgumentException("The input stream cannot be read.");
            }
            else if(!input.CanSeek)
            {
                throw new ArgumentException("The input stream does not support seeking.");
            }

            #endregion

            this.input = input;

            input.Position = 0;

            if(IsRiffFile())
            {
                // Skip RIFF length.
                input.ReadByte();
                input.ReadByte();
                input.ReadByte();
                input.ReadByte();

                waveChunkPosition = FindChunk(new byte[] { (byte)'W', (byte)'A', (byte)'V', (byte)'E' });

                if(waveChunkPosition < 0)
                {
                    throw new ChunkNotFoundException("WAVE");
                }

                if(FindChunk(new byte[] { (byte)'f', (byte)'m', (byte)'t', (byte)' ' }) < 0)
                {
                    throw new ChunkNotFoundException("fmt ");
                }

                ParseFormatData();

                // Reposition stream to the beginning of the wave chunk.
                input.Position = waveChunkPosition;

                if(FindChunk(new byte[] { (byte)'d', (byte)'a', (byte)'t', (byte)'a' }) < 0)
                {
                    throw new ChunkNotFoundException("data");
                }

                ParseWaveData();
            }
            else
            {
                throw new ChunkNotFoundException("RIFF");
            }
        }

        #endregion

        #region Methods

        #region Helper Methods

        private bool IsRiffFile()
        {
            bool result;

            if(input.ReadByte() == (byte)'R' &&
                input.ReadByte() == (byte)'I' &&
                input.ReadByte() == (byte)'F' &&
                input.ReadByte() == (byte)'F')
            {
                result = true;
            }
            else
            {
                result = false;
            }

            return result;
        }

        private long FindChunk(byte[] chunkID)
        {
            Debug.Assert(chunkID != null && chunkID.Length == ChunkIDLength);

            bool found = false;
            int b;
            int i = 0;

            // "Prime the pump."
            b = input.ReadByte();

            while(!found && b >= 0)
            {
                if(b == chunkID[i])
                {
                    i++;

                    if(i < chunkID.Length)
                    {
                        b = input.ReadByte();
                    }
                    else
                    {
                        found = true;
                    }
                }
                else
                {
                    b = input.ReadByte();
                    i = 0;
                }
            }

            return (found ? input.Position : -1);
        }

        private void ParseFormatData()
        {
            // Skip format length.
            input.ReadByte();
            input.ReadByte();
            input.ReadByte();
            input.ReadByte();

            byte[] shortArray = new byte[2];
            byte[] intArray = new byte[4];

            input.Read(shortArray, 0, shortArray.Length);
            short formatTag = BitConverter.ToInt16(shortArray, 0);

            if(formatTag != 1)
            {
                // If the format is not PCM/Uncompressed, throw an exception.
                throw new NotSupportedException("Non PCM/Uncompressed formats are not supported at this time.");
            }

            input.Read(shortArray, 0, shortArray.Length);
            channelCount = BitConverter.ToInt16(shortArray, 0);

            input.Read(intArray, 0, intArray.Length);
            samplesPerSecond = BitConverter.ToInt32(intArray, 0);

            // Skip average bytes per second.
            input.Read(intArray, 0, intArray.Length);

            // Skip block align.
            input.Read(shortArray, 0, shortArray.Length);

            input.Read(shortArray, 0, shortArray.Length);
            bitsPerSecond = BitConverter.ToInt16(shortArray, 0);
        }

        private void ParseWaveData()
        {
            byte[] uintArray = new byte[4];

            input.Read(uintArray, 0, uintArray.Length);
            bytesLeft = BitConverter.ToUInt32(uintArray, 0);
            waveDataLength = bytesLeft;
        }

        #endregion

        public byte Read()
        {
            #region Require

            if(disposed)
            {
                throw new ObjectDisposedException(this.GetType().Name);
            }
            else if(bytesLeft == 0)
            {
                throw new EndOfStreamException();
            }

            #endregion

            bytesLeft--;

            int b = input.ReadByte();

            Debug.Assert(b >= 0);

            return (byte)b;
        }

        public byte[] Read(int count)
        {
            #region Require

            if(disposed)
            {
                throw new ObjectDisposedException(this.GetType().Name);
            }
            else if(count < 0)
            {
                throw new ArgumentOutOfRangeException("count");
            }

            #endregion

            byte[] buffer;

            if(count <= bytesLeft)
            {
                buffer = new byte[count];

                bytesLeft -= count;
            }
            else
            {
                buffer = new byte[bytesLeft];

                bytesLeft = 0;
            }

            input.Read(buffer, 0, buffer.Length);

            return buffer;
        }

        public void Close()
        {
            Dispose();
        }

        #endregion

        #region Properties

        public Stream BaseStream
        {
            get
            {
                return input;
            }
        }

        public short ChannelCount
        {
            get
            {
                return channelCount;
            }
        }

        public int SamplesPerSecond
        {
            get
            {
                return samplesPerSecond;
            }
        }

        public short BitsPerSecond
        {
            get
            {
                return bitsPerSecond;
            }
        }

        public long WaveDataLength
        {
            get
            {
                return waveDataLength;
            }
        }

        #endregion

        #endregion

        #region IDisposable Members

        public void Dispose()
        {
            #region Guard

            if(disposed)
            {
                return;
            }

            #endregion

            input.Close();

            disposed = true;

            GC.SuppressFinalize(this);
        }

        #endregion
    }
}
