#region License

/* Copyright (c) 2007 Leslie Sanford
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to 
 * deal in the Software without restriction, including without limitation the 
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
 * sell copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software. 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
 * THE SOFTWARE.
 */

#endregion

#region Contact

/*
 * Leslie Sanford
 * Email: jabberdabber@hotmail.com
 */

#endregion

using System;
using System.Diagnostics;
using System.IO;

namespace Sanford.Multimedia.Wave
{
    /// <summary>
    /// 
    /// </summary>
    public sealed class WaveWriter : IDisposable
    {
        #region WaveWriter Members

        #region Fields

        #region Constants

        private const int FormatChunkSize = 16;

        private const int RiffLengthOffset = 4;

        private const int DataLengthOffset = 40;

        #endregion

        private Stream waveStream;

        private short blockAlign;

        private int averageBytesPerSecond;

        private bool disposed = false;

        #endregion

        #region Construction

        /// <summary>
        /// Initializes a new instance of the WaveWriter class that writes to a stream.
        /// </summary>
        /// <param name="channelCount">
        /// The number of waveform channels.
        /// </param>
        /// <param name="samplesPerSecond">
        /// The sample rate.
        /// </param>
        /// <param name="bitsPerSample">
        /// The bit resolution.
        /// </param>
        /// <exception cref="ArgumentOutOfRangeException">
        /// channelCount is less than one, or samplesPerSecond is less than one, or bitsPerSample is less than one.
        /// </exception>       
        public WaveWriter(short channelCount, int samplesPerSecond, short bitsPerSample)
        {
            Initialize(new MemoryStream(), channelCount, samplesPerSecond, bitsPerSample);
        }

        /// <summary>
        /// Initializes a new instance of the WaveWriter class based on the supplied stream. 
        /// </summary>
        /// <param name="output">
        /// The output stream.
        /// </param>
        /// <param name="channelCount">
        /// The number of waveform channels.
        /// </param>
        /// <param name="samplesPerSecond">
        /// The sample rate.
        /// </param>
        /// <param name="bitsPerSample">
        /// The bit resolution.
        /// </param>
        /// <exception cref="ArgumentNullException">
        /// output is a null reference.
        /// </exception>
        /// <exception cref="ArgumentException">
        /// output does not support writing, seeking, is already closed, or is not empty.
        /// </exception>
        /// <exception cref="ArgumentOutOfRangeException">
        /// channelCount is less than one, or samplesPerSecond is less than one, or bitsPerSample is less than one.
        /// </exception>   
        public WaveWriter(Stream output, short channelCount, int samplesPerSecond, short bitsPerSample)
        {
            Initialize(output, channelCount, samplesPerSecond, bitsPerSample);
        }

        #endregion

        #region Methods

        #region Helper Methods

        private void Initialize(Stream output, short channelCount, int samplesPerSecond, short bitsPerSample)
        {
            #region Require

            if(output == null)
            {
                throw new ArgumentNullException("output");
            }
            else if(!output.CanWrite)
            {
                throw new ArgumentException("The output stream cannot be written to.");
            }
            else if(!output.CanSeek)
            {
                throw new ArgumentException("The stream does not support seeking.");
            }
            else if(output.Length != 0)
            {
                throw new ArgumentException("The stream is not empty.");
            }
            else if(channelCount < 1)
            {
                throw new ArgumentOutOfRangeException("channelCount");
            }
            else if(samplesPerSecond < 1)
            {
                throw new ArgumentOutOfRangeException("samplesPerSecond");
            }
            else if(bitsPerSample < 1)
            {
                throw new ArgumentOutOfRangeException("bitsPerSample");
            }

            #endregion

            this.waveStream = output;

            CalculateBlockAlign(channelCount, bitsPerSample);
            CalcuateAverageBytesPerSecond(samplesPerSecond);

            WriteRiffChunk(channelCount, samplesPerSecond, bitsPerSample);
        }

        private void CalculateBlockAlign(short channelCount, short bitsPerSample)
        {
            if(bitsPerSample % 8 == 0)
            {
                blockAlign = (short)(bitsPerSample / 8 * channelCount);
            }
            else
            {
                blockAlign = (short)((bitsPerSample / 8 + 1) * channelCount);
            }
        }

        private void CalcuateAverageBytesPerSecond(int samplesPerSecond)
        {
            averageBytesPerSecond = blockAlign * samplesPerSecond;
        }

        private void WriteRiffChunk(short channelCount, int samplesPerSecond, short bitsPerSample)
        {
            WriteRiffChunkID();

            // Placeholder for wave file length.
            waveStream.Write(BitConverter.GetBytes(0), 0, 4);

            WriteWaveChunk(channelCount, samplesPerSecond, bitsPerSample);
        }

        private void WriteRiffChunkID()
        {
            waveStream.WriteByte((byte)'R');
            waveStream.WriteByte((byte)'I');
            waveStream.WriteByte((byte)'F');
            waveStream.WriteByte((byte)'F');
        }

        private void WriteWaveChunk(short channelCount, int samplesPerSecond, short bitsPerSample)
        {
            WriteWaveChunkID();
            WriteFormatChunk(channelCount, samplesPerSecond, bitsPerSample);
            WriteDataChunk();
        }

        private void WriteWaveChunkID()
        {
            waveStream.WriteByte((byte)'W');
            waveStream.WriteByte((byte)'A');
            waveStream.WriteByte((byte)'V');
            waveStream.WriteByte((byte)'E');
        }

        private void WriteFormatChunk(short channelCount, int samplesPerSecond, short bitsPerSample)
        {
            WriteFormatChunkID();

            waveStream.Write(BitConverter.GetBytes(FormatChunkSize), 0, 4);
            waveStream.Write(BitConverter.GetBytes((short)1), 0, 2); // Format type (PCM/Uncompressed).
            waveStream.Write(BitConverter.GetBytes(channelCount), 0, 2);
            waveStream.Write(BitConverter.GetBytes(samplesPerSecond), 0, 4);
            waveStream.Write(BitConverter.GetBytes(averageBytesPerSecond), 0, 4);
            waveStream.Write(BitConverter.GetBytes(blockAlign), 0, 2);
            waveStream.Write(BitConverter.GetBytes(bitsPerSample), 0, 2);
        }

        private void WriteFormatChunkID()
        {
            waveStream.WriteByte((byte)'f');
            waveStream.WriteByte((byte)'m');
            waveStream.WriteByte((byte)'t');
            waveStream.WriteByte((byte)' ');
        }

        private void WriteDataChunk()
        {
            WriteDataChunkID();

            // Placeholder for wave data length.
            waveStream.Write(BitConverter.GetBytes(0), 0, 4);
        }

        private void WriteDataChunkID()
        {
            waveStream.WriteByte((byte)'d');
            waveStream.WriteByte((byte)'a');
            waveStream.WriteByte((byte)'t');
            waveStream.WriteByte((byte)'a');
        }

        private void PadWaveData()
        {
            if(waveStream.Length % 2 != 0)
            {
                waveStream.Seek(waveStream.Length, SeekOrigin.Begin);

                // If data length is not word aligned, write an extra byte.
                Write(0);
            }

            Debug.Assert(waveStream.Length % 2 == 0, "Length not word aligned.");
        }

        #endregion

        /// <summary>
        /// Writes an unsigned byte to the current stream and advances the stream position by one byte.
        /// </summary>
        /// <param name="sample">
        /// The unsigned byte to write.
        /// </param>
        /// <exception cref="IOException">
        /// An IO error occurs.
        /// </exception>
        /// <exception cref="ObjectDisposedException">
        /// The stream is closed.
        /// </exception>
        public void Write(byte sample)
        {
            #region Require

            if(disposed)
            {
                throw new ObjectDisposedException(this.GetType().Name);
            }

            #endregion

            waveStream.WriteByte(sample);
        }

        /// <summary>
        /// Writes a byte array to the underlying stream.
        /// </summary>
        /// <param name="samples">
        /// A byte array containing the waveform data to write.
        /// </param>
        /// <exception cref="IOException">
        /// An IO error occurs.
        /// </exception>
        /// <exception cref="ObjectDisposedException">
        /// The stream is closed.
        /// </exception>
        /// <exception cref="ArgumentNullException">
        /// samples is a null reference.
        /// </exception>
        public void Write(byte[] samples)
        {
            #region Require

            if(disposed)
            {
                throw new ObjectDisposedException(this.GetType().Name);
            }
            else if(samples == null)
            {
                throw new ArgumentNullException("samples");
            }

            #endregion

            waveStream.Write(samples, 0, samples.Length);
        }

        /// <summary>
        /// Closes the current WaveWriter and the underlying stream.
        /// </summary>
        public void Close()
        {
            Dispose();
        }

        #endregion

        #region Properties

        /// <summary>
        /// Gets the underlying stream of the WaveWriter.
        /// </summary>
        public Stream BaseStream
        {
            get
            {
                return waveStream;
            }
        }

        #endregion

        #endregion

        #region IDisposable Members

        public void Dispose()
        {
            #region Guard

            if(disposed)
            {
                return;
            }

            #endregion            

            // Seek to the RIFF length field.
            waveStream.Seek(RiffLengthOffset, SeekOrigin.Begin);

            // Write the length of the wave file not counting the RIFF chunk 
            // ID and the RIFF length field.
            waveStream.Write(BitConverter.GetBytes(waveStream.Length - (RiffLengthOffset + 4)), 0, 4);

            // Seek to the data length field.
            waveStream.Seek(DataLengthOffset, SeekOrigin.Begin);

            // Write the length of the data chunk not counting the data chunk 
            // ID and the data length field.
            waveStream.Write(BitConverter.GetBytes(waveStream.Length - (DataLengthOffset + 4)), 0, 4);

            // Ensure that the wave file is word aligned - this step should be last before closing.
            PadWaveData();

            waveStream.Close();

            disposed = true;

            GC.SuppressFinalize(this);
        }

        #endregion
    }
}
